@@ -68,17 +68,19 @@ extern "C" {
 #include "srl_reader_varint.h"
 #include "srl_reader_decompress.h"
 
-#define srl_stack_push_and_set(iter, tag, length) STMT_START {  \
+#define srl_stack_push_and_set(_iter, _tag, _length) STMT_START {  \
     srl_iterator_stack_ptr _stack_ptr;                          \
-    srl_stack_push_ptr((iter)->stack, _stack_ptr);              \
-    _stack_ptr->offset = SRL_RDR_BODY_POS_OFS((iter->pbuf));    \
-    _stack_ptr->count = (length);                               \
-    _stack_ptr->idx = (length);                                 \
-    _stack_ptr->tag = (tag);                                    \
+    srl_stack_push_ptr((_iter)->stack, _stack_ptr);             \
+    _stack_ptr->offset = SRL_RDR_BODY_POS_OFS((_iter)->pbuf);   \
+    _stack_ptr->count = (_length);                              \
+    _stack_ptr->idx = (_length);                                \
+    _stack_ptr->tag = (_tag);                                   \
 } STMT_END
 
 #define srl_stack_type_t srl_iterator_stack_t
 #include "srl_stack.h"
+#define SRL_ITER_STACK_ROOT_TAG SRL_HDR_PACKET_START
+#define SRL_ITER_STACK_ON_ROOT(stack) ((stack)->ptr->tag == SRL_ITER_STACK_ROOT_TAG)
 
 #define SRL_ITER_NOT_FOUND (0) /* 0 is invalid offset */
 #define SRL_ITER_BASE_ERROR_FORMAT              "Sereal::Path::Iterator: Error in %s:%u "
@@ -87,12 +89,13 @@ extern "C" {
 #define SRL_ITER_ERROR(msg)                     croak(SRL_ITER_BASE_ERROR_FORMAT "%s", SRL_ITER_BASE_ERROR_ARGS, (msg))
 #define SRL_ITER_ERRORf1(fmt, var)              croak(SRL_ITER_BASE_ERROR_FORMAT fmt,  SRL_ITER_BASE_ERROR_ARGS, (var))
 #define SRL_ITER_ERRORf2(fmt, var1, var2)       croak(SRL_ITER_BASE_ERROR_FORMAT fmt,  SRL_ITER_BASE_ERROR_ARGS, (var1), (var2))
+#define SRL_ITER_ERRORf3(fmt, var1, var2, var3) croak(SRL_ITER_BASE_ERROR_FORMAT fmt,  SRL_ITER_BASE_ERROR_ARGS, (var1), (var2), (var3))
 
 #ifdef TRACE_ITERATOR
-#   define SRL_ITER_TRACE(msg, args...)            SRL_RDR_TRACE("%s" msg, srl_debug_tabulator(iter), ## args)
+#   define SRL_ITER_TRACE(msg, args...)            SRL_RDR_TRACE("%s " msg, srl_debug_tabulator(iter), ## args)
 #   define SRL_ITER_REPORT_TAG(iter, tag) STMT_START {                              \
         SRL_RDR_TRACE(                                                              \
-            "%stag SRL_HDR_%s (int: %d hex: 0x%x) at ofs %"UVuf" body_ofs %"UVuf,   \
+            "%s tag SRL_HDR_%s (int: %d hex: 0x%x) at ofs %"UVuf" body_ofs %"UVuf,  \
             srl_debug_tabulator((iter)),                                            \
             SRL_TAG_NAME((tag)),                                                    \
             (tag), (tag),                                                           \
@@ -100,9 +103,23 @@ extern "C" {
             (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf)                                 \
         );                                                                          \
     } STMT_END
+#   define SRL_ITER_REPORT_STACK_STATE(iter) STMT_START {                           \
+        srl_iterator_stack_ptr stack_ptr = (iter)->stack->ptr;                      \
+        SRL_RDR_TRACE(                                                              \
+            "%s stack state depth=%"IVdf" tag=SRL_HDR_%s (int: %d hex: 0x%x) "      \
+            "count=%u idx=%d offset=%"UVuf,                                         \
+            srl_debug_tabulator((iter)),                                            \
+            SRL_STACK_DEPTH((iter)->stack),                                         \
+            SRL_TAG_NAME(stack_ptr->tag), stack_ptr->tag, stack_ptr->tag,           \
+            stack_ptr->count,                                                       \
+            stack_ptr->idx,                                                         \
+            stack_ptr->offset                                                       \
+        );                                                                          \
+    } STMT_END
 #else
 #   define SRL_ITER_TRACE(msg, args...)
 #   define SRL_ITER_REPORT_TAG(iter, tag)
+#   define SRL_ITER_REPORT_STACK_STATE(iter)
 #endif
 
 #define SRL_ITER_ASSERT_EOF(iter, msg) STMT_START {                                 \
@@ -112,10 +129,10 @@ extern "C" {
 } STMT_END
 
 #define SRL_ITER_ASSERT_STACK(iter) STMT_START {                                    \
-    if (expect_false(srl_stack_empty((iter)->stack))) {                             \
+    assert(!srl_stack_empty((iter)->stack));                                        \
+    if (expect_false(((int) (iter)->stack->ptr->idx)) < 0) {                        \
         SRL_ITER_ERROR("Stack is empty! Inconsistent state!");                      \
     }                                                                               \
-    assert((iter)->stack->ptr->idx >= 0);                                           \
 } STMT_END
 
 #define SRL_ITER_ASSERT_ARRAY_ON_STACK(iter) STMT_START {                           \
@@ -143,8 +160,7 @@ extern "C" {
 } STMT_END
 
 /* function declaration */
-SRL_STATIC_INLINE void srl_iterator_step_internal(pTHX_ srl_iterator_t *iter);
-SRL_STATIC_INLINE void srl_iterator_wrap_stack(pTHX_ srl_iterator_t *iter, IV depth);
+SRL_STATIC_INLINE void srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter);
 
 /* wrappers */
 UV srl_iterator_eof(pTHX_ srl_iterator_t *iter)     { return SRL_RDR_DONE(iter->pbuf) ? 1 : 0; }
@@ -155,7 +171,7 @@ srl_debug_tabulator(pTHX_ srl_iterator_t *iter)
 {
     int i = 0;
     static char buf[1024];
-    for (; i < SRL_STACK_DEPTH(iter->stack); ++i) buf[i] = '-';
+    for (; i < SRL_STACK_DEPTH(iter->stack); ++i) buf[i] = '+';
     buf[i] = '\0';
     return (const char *) buf; // XXX
 }
@@ -178,7 +194,7 @@ srl_build_iterator_struct(pTHX_ HV *opt)
 
     // TODO inline stack
     // TODO keep fixed stack size ???
-    if (expect_false(srl_stack_init(aTHX_ stack, 1024) != 0)) {
+    if (expect_false(srl_stack_init(aTHX_ stack, 32) != 0)) {
         Safefree(iter);
         Safefree(stack);
         croak("Out of memory");
@@ -189,7 +205,6 @@ srl_build_iterator_struct(pTHX_ HV *opt)
     iter->stack = stack;
     iter->document = NULL;
     iter->tmp_buf_owner = NULL;
-    iter->first_tag_offset = 0;
     iter->dec = NULL;
 
     /* load options */
@@ -290,168 +305,228 @@ srl_iterator_set(pTHX_ srl_iterator_t *iter, SV *src)
     SRL_RDR_UPDATE_BODY_POS(iter->pbuf, protocol_version);
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 
-    iter->first_tag_offset = SRL_RDR_BODY_POS_OFS(iter->pbuf);
+    srl_stack_push_and_set(iter, SRL_ITER_STACK_ROOT_TAG, 1);
     srl_iterator_reset(aTHX_ iter);
 }
 
 void
 srl_iterator_reset(pTHX_ srl_iterator_t *iter)
-{
-    U8 tag = '\0';
-    SRL_ITER_TRACE();
-    srl_stack_clear(iter->stack);
-    srl_stack_push_and_set(iter, tag, 1);
-    iter->buf.pos = iter->buf.body_pos + iter->first_tag_offset;
-    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
-}
-
-SRL_STATIC_INLINE void
-srl_iterator_wrap_stack(pTHX_ srl_iterator_t *iter, IV expected_depth)
 {
     srl_stack_t *stack = iter->stack;
-    if (expect_false(SRL_STACK_DEPTH(stack) < expected_depth))
-        SRL_ITER_ERRORf2("expected depth %"IVdf" is higher than current depth %"IVdf,
-                         expected_depth, SRL_STACK_DEPTH(stack));
+    SRL_ITER_TRACE("--------------------------");
 
-    while (!srl_stack_empty(stack) && stack->ptr->idx == 0) {
-        if (SRL_STACK_DEPTH(stack) == expected_depth) break;
-        srl_stack_pop_nocheck(stack);
-        if (srl_stack_empty(stack))
-            SRL_ITER_TRACE("end of stack reached");
-    }
-}
-
-/* Main routine. Caller must ensure that EOF is NOT reached */
-SRL_STATIC_INLINE void
-srl_iterator_step_internal(pTHX_ srl_iterator_t *iter)
-{
-    U8 tag;
-    UV length;
-    srl_stack_t *stack = iter->stack;
-
-    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
-
-    srl_iterator_wrap_stack(aTHX_ iter, -1);
-    if (srl_stack_empty(stack)) return;
-
-    SRL_ITER_ASSERT_STACK(iter);
-
-    stack->ptr->idx--;
-    SRL_ITER_TRACE("stack->ptr: idx=%d depth=%d",
-                   stack->ptr->idx, (int) SRL_STACK_DEPTH(stack));
-
-    SRL_ITER_ASSERT_STACK(iter);
-
-read_again:
-    tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;
-    SRL_ITER_REPORT_TAG(iter, tag);
-    iter->buf.pos++;
-
-    /* No code which decrease step, next or stack's counters should be added here.
-     * Otherwise the counters will be decreased twicer for tags like REFN, ALIAS, etc. */
-
-    switch (tag) {
-        CASE_SRL_HDR_SHORT_BINARY:
-            iter->buf.pos += SRL_HDR_SHORT_BINARY_LEN_FROM_TAG(tag);
-            break;
-
-        case SRL_HDR_HASH:
-            length = srl_read_varint_uv_count(aTHX_ iter->pbuf, " while reading HASH");
-            if (length > 0) srl_stack_push_and_set(iter, tag, length * 2);
-            break;
-
-        case SRL_HDR_ARRAY:
-            length = srl_read_varint_uv_count(aTHX_ iter->pbuf, " while reading ARRAY");
-            if (length > 0) srl_stack_push_and_set(iter, tag, length);
-            break;
-
-        CASE_SRL_HDR_HASHREF:
-            length = SRL_HDR_HASHREF_LEN_FROM_TAG(tag);
-            if (length > 0) srl_stack_push_and_set(iter, tag, length * 2);
-            break;
-
-        CASE_SRL_HDR_ARRAYREF:
-            length = SRL_HDR_ARRAYREF_LEN_FROM_TAG(tag);
-            if (length > 0) srl_stack_push_and_set(iter, tag, length);
-            break;
-
-        CASE_SRL_HDR_POS:
-        CASE_SRL_HDR_NEG:
-            break;
-
-        case SRL_HDR_VARINT:
-        case SRL_HDR_ZIGZAG:
-            srl_skip_varint(aTHX_ iter->pbuf);
-            break;
-
-        case SRL_HDR_FLOAT:         iter->buf.pos += 4;      break;
-        case SRL_HDR_DOUBLE:        iter->buf.pos += 8;      break;
-        case SRL_HDR_LONG_DOUBLE:   iter->buf.pos += 16;     break;
-
-        case SRL_HDR_TRUE:
-        case SRL_HDR_FALSE:
-        case SRL_HDR_UNDEF:
-        case SRL_HDR_CANONICAL_UNDEF:
-            break;
-
-        case SRL_HDR_REFN:
-        case SRL_HDR_ALIAS:
-        case SRL_HDR_WEAKEN:
-            goto read_again;
-
-        case SRL_HDR_PAD:
-            while (SRL_RDR_NOT_DONE(iter->pbuf) && *iter->buf.pos++ == SRL_HDR_PAD) {};
-            goto read_again;
-
-        case SRL_HDR_BINARY:
-        case SRL_HDR_STR_UTF8:
-            length = srl_read_varint_uv_length(aTHX_ iter->pbuf, " while reading BINARY or STR_UTF8");
-            iter->buf.pos += length;
-            break;
-
-        case SRL_HDR_COPY:
-        case SRL_HDR_REFP:
-            srl_skip_varint(aTHX_ iter->pbuf);
-            break;
-
-        /* case SRL_HDR_OBJECTV: */
-        /* case SRL_HDR_OBJECTV_FREEZE: */
-        /* case SRL_HDR_REGEXP: */
-        /* case SRL_HDR_OBJECT: */
-        /* case SRL_HDR_OBJECT_FREEZE: */
-
-        default:
-            SRL_RDR_ERROR_UNIMPLEMENTED(iter->pbuf, tag, "");
-            break;
+    while (!SRL_ITER_STACK_ON_ROOT(stack)) {
+        srl_stack_pop(stack); // does empty check internally
     }
 
+    srl_iterator_restore_stack_position(aTHX_ iter);
+}
+
+void
+srl_iterator_unite(pTHX_ srl_iterator_t *iter)
+{
+    UV offset;
+    srl_stack_t *stack = iter->stack;
+    SRL_ITER_TRACE("--------------------------");
+
+    if (expect_false(SRL_STACK_DEPTH(stack) <= 0))
+        SRL_ITER_ERROR("There is nothing to unite. Please call disjoin first.");
+
+    while (!SRL_ITER_STACK_ON_ROOT(stack)) {
+        srl_stack_pop(stack);
+    }
+
+    offset = stack->ptr->offset;
+    srl_stack_pop(stack); // remove SRL_ITER_STACK_ROOT_TAG
+
+    SRL_ITER_ASSERT_STACK(iter);
+    iter->buf.pos = iter->buf.body_pos + offset;
+
+    SRL_ITER_REPORT_STACK_STATE(iter);
+    SRL_ITER_TRACE("ofs %"UVuf" body_ofs %"UVuf,
+                   (UV) SRL_RDR_POS_OFS((iter)->pbuf),
+                   (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf));
+
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 }
 
+void
+srl_iterator_disjoin(pTHX_ srl_iterator_t *iter)
+{
+    srl_iterator_stack_ptr stack_ptr;
+
+    SRL_ITER_REPORT_STACK_STATE(iter);
+    SRL_ITER_TRACE("ofs %"UVuf" body_ofs %"UVuf,
+                   (UV) SRL_RDR_POS_OFS((iter)->pbuf),
+                   (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf));
+
+    // This record apart of being a boundary stores offset to idx's tag (i.e
+    // current tag). By default stack keeps offset to tag's starting point
+    // (i.e. continer located in the buf).
+
+    srl_stack_push_ptr(iter->stack, stack_ptr);
+    stack_ptr->offset = SRL_RDR_BODY_POS_OFS(iter->pbuf); // disjoint point
+    stack_ptr->tag = SRL_ITER_STACK_ROOT_TAG;
+    stack_ptr->count = 1;
+    stack_ptr->idx = 1;
+}
+
+
+SRL_STATIC_INLINE void
+srl_iterator_restore_stack_position(pTHX_ srl_iterator_t *iter)
+{
+    srl_stack_t *stack = iter->stack;
+    SRL_ITER_ASSERT_STACK(iter);
+
+    stack->ptr->idx = stack->ptr->count;
+    iter->buf.pos = iter->buf.body_pos + stack->ptr->offset;
+
+    SRL_ITER_REPORT_STACK_STATE(iter);
+    SRL_ITER_TRACE("ofs %"UVuf" body_ofs %"UVuf,
+                   (UV) SRL_RDR_POS_OFS((iter)->pbuf),
+                   (UV) SRL_RDR_BODY_POS_OFS((iter)->pbuf));
+
+    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
+}
+
+#define srl_iterator_wrap_stack(iter, expected_depth, stack_ptr) STMT_START {                   \
+    srl_stack_t *stack = (iter)->stack;                                                         \
+    (stack_ptr) = stack->ptr;                                                                   \
+                                                                                                \
+    SRL_ITER_TRACE("expected_depth=%"IVdf, (expected_depth));                                   \
+    SRL_ITER_REPORT_STACK_STATE(iter);                                                          \
+                                                                                                \
+    assert(((expected_depth) > (IV) SRL_STACK_DEPTH(stack)) == 0);                              \
+                                                                                                \
+    while ((stack_ptr)->idx == 0 && (stack_ptr)->tag != SRL_ITER_STACK_ROOT_TAG) {              \
+        if (SRL_STACK_DEPTH(stack) == (expected_depth)) break;                                  \
+        srl_stack_pop_nocheck(stack);                                                           \
+        (stack_ptr) = stack->ptr;                                                               \
+        if (SRL_ITER_STACK_ON_ROOT(stack))                                                      \
+            SRL_ITER_TRACE("root of stack reached");                                            \
+    }                                                                                           \
+} STMT_END
+
+/* Main routine. Caller MUST ensure that EOF is NOT reached */
+/* and MUST call srl_iterator_wrap_stack before calling srl_iterator_step_internal() */
+#define srl_iterator_step_internal(iter, stack_ptr) STMT_START {                                \
+    U8 tag;                                                                                     \
+    UV length;                                                                                  \
+                                                                                                \
+    if (expect_false((stack_ptr)->idx == 0))                                                    \
+        SRL_ITER_ERRORf1("Nothing to parse at depth=%"IVdf, SRL_STACK_DEPTH(iter->stack));      \
+                                                                                                \
+    (stack_ptr)->idx--;                                                                         \
+                                                                                                \
+    SRL_ITER_ASSERT_STACK(iter);                                                                \
+    DEBUG_ASSERT_RDR_SANE(iter->pbuf);                                                          \
+                                                                                                \
+read_again:                                                                                     \
+    tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;                                                 \
+    SRL_ITER_REPORT_TAG(iter, tag);                                                             \
+    iter->buf.pos++;                                                                            \
+                                                                                                \
+    /* No code which decrease step, next or stack's counters should be added here.              \
+     * Otherwise the counters will be decreased twicer for tags like REFN, ALIAS, etc. */       \
+                                                                                                \
+    switch (tag) {                                                                              \
+        CASE_SRL_HDR_SHORT_BINARY:                                                              \
+            iter->buf.pos += SRL_HDR_SHORT_BINARY_LEN_FROM_TAG(tag);                            \
+            break;                                                                              \
+                                                                                                \
+        case SRL_HDR_HASH:                                                                      \
+            length = srl_read_varint_uv_count(aTHX_ iter->pbuf, " while reading HASH");         \
+            if (length > 0) srl_stack_push_and_set(iter, tag, length * 2);                      \
+            break;                                                                              \
+                                                                                                \
+        case SRL_HDR_ARRAY:                                                                     \
+            length = srl_read_varint_uv_count(aTHX_ iter->pbuf, " while reading ARRAY");        \
+            if (length > 0) srl_stack_push_and_set(iter, tag, length);                          \
+            break;                                                                              \
+                                                                                                \
+        CASE_SRL_HDR_HASHREF:                                                                   \
+            length = SRL_HDR_HASHREF_LEN_FROM_TAG(tag);                                         \
+            if (length > 0) srl_stack_push_and_set(iter, tag, length * 2);                      \
+            break;                                                                              \
+                                                                                                \
+        CASE_SRL_HDR_ARRAYREF:                                                                  \
+            length = SRL_HDR_ARRAYREF_LEN_FROM_TAG(tag);                                        \
+            if (length > 0) srl_stack_push_and_set(iter, tag, length);                          \
+            break;                                                                              \
+                                                                                                \
+        CASE_SRL_HDR_POS:                                                                       \
+        CASE_SRL_HDR_NEG:                                                                       \
+            break;                                                                              \
+                                                                                                \
+        case SRL_HDR_VARINT:                                                                    \
+        case SRL_HDR_ZIGZAG:                                                                    \
+            srl_skip_varint(aTHX_ iter->pbuf);                                                  \
+            break;                                                                              \
+                                                                                                \
+        case SRL_HDR_FLOAT:         iter->buf.pos += 4;      break;                             \
+        case SRL_HDR_DOUBLE:        iter->buf.pos += 8;      break;                             \
+        case SRL_HDR_LONG_DOUBLE:   iter->buf.pos += 16;     break;                             \
+                                                                                                \
+        case SRL_HDR_TRUE:                                                                      \
+        case SRL_HDR_FALSE:                                                                     \
+        case SRL_HDR_UNDEF:                                                                     \
+        case SRL_HDR_CANONICAL_UNDEF:                                                           \
+            break;                                                                              \
+                                                                                                \
+        case SRL_HDR_REFN:                                                                      \
+        case SRL_HDR_ALIAS:                                                                     \
+        case SRL_HDR_WEAKEN:                                                                    \
+            goto read_again;                                                                    \
+                                                                                                \
+        case SRL_HDR_PAD:                                                                       \
+            while (SRL_RDR_NOT_DONE(iter->pbuf) && *iter->buf.pos++ == SRL_HDR_PAD) {};         \
+            goto read_again;                                                                    \
+                                                                                                \
+        case SRL_HDR_BINARY:                                                                    \
+        case SRL_HDR_STR_UTF8:                                                                  \
+            length = srl_read_varint_uv_length(aTHX_ iter->pbuf,                                \
+                                              " while reading BINARY or STR_UTF8");             \
+            iter->buf.pos += length;                                                            \
+            break;                                                                              \
+                                                                                                \
+        case SRL_HDR_COPY:                                                                      \
+        case SRL_HDR_REFP:                                                                      \
+            srl_skip_varint(aTHX_ iter->pbuf);                                                  \
+            break;                                                                              \
+                                                                                                \
+        /* case SRL_HDR_OBJECTV: */                                                             \
+        /* case SRL_HDR_OBJECTV_FREEZE: */                                                      \
+        /* case SRL_HDR_REGEXP: */                                                              \
+        /* case SRL_HDR_OBJECT: */                                                              \
+        /* case SRL_HDR_OBJECT_FREEZE: */                                                       \
+                                                                                                \
+        default:                                                                                \
+            SRL_RDR_ERROR_UNIMPLEMENTED(iter->pbuf, tag, "");                                   \
+            break;                                                                              \
+    }                                                                                           \
+                                                                                                \
+    DEBUG_ASSERT_RDR_SANE(iter->pbuf);                                                          \
+} STMT_END
+
 /* srl_iterator_step_in() does N steps. Where step is a serialized object */
 
 void
 srl_iterator_step_in(pTHX_ srl_iterator_t *iter, UV n)
 {
-    srl_stack_t *stack = iter->stack;
+    srl_iterator_stack_ptr stack_ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
-    SRL_ITER_TRACE("n=%"UVuf, n);
-
     SRL_ITER_ASSERT_STACK(iter);
-    if (expect_false(n == 0)) return;
 
-    while (expect_true(!srl_stack_empty(stack))) {
-        if (n == 0) break;
-        srl_iterator_step_internal(aTHX_ iter);
-        n--;
+    SRL_ITER_TRACE("n=%"UVuf, n);
+    SRL_ITER_REPORT_STACK_STATE(iter);
+
+    while (n--) {
+        srl_iterator_wrap_stack(iter, -1, stack_ptr);
+        srl_iterator_step_internal(iter, stack_ptr);
     }
 
-    if (expect_false(n != 0)) {
-        SRL_ITER_ERRORf1("Failed to do %"UVuf" steps. Likely EOF was reached", n);
-    }
-
-    SRL_ITER_TRACE("Did expected number of steps");
+    SRL_ITER_TRACE("Completed expected number of steps");
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 }
 
@@ -462,24 +537,26 @@ void
 srl_iterator_next(pTHX_ srl_iterator_t *iter, UV n)
 {
     srl_stack_t *stack = iter->stack;
+    srl_iterator_stack_ptr stack_ptr;
     IV expected_depth = SRL_STACK_DEPTH(stack);
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
-    SRL_ITER_TRACE("n=%"UVuf, n);
-
     SRL_ITER_ASSERT_STACK(iter);
-    if (expect_false(n == 0)) return;
-    if (expect_false(stack->ptr->idx == 0))
-        SRL_ITER_ERROR("Nothing to parse at this depth");
 
-    while (expect_true(!srl_stack_empty(stack))) {
+    SRL_ITER_TRACE("n=%"UVuf, n);
+    SRL_ITER_REPORT_STACK_STATE(iter);
+
+    if (expect_false(n == 0)) return;
+
+    while (1) {
+        srl_iterator_wrap_stack(iter, expected_depth, stack_ptr);
+
         if (SRL_STACK_DEPTH(stack) == expected_depth) {
             if (n == 0) break;
             else n--;
         }
 
-        srl_iterator_step_internal(aTHX_ iter);
-        srl_iterator_wrap_stack(aTHX_ iter, expected_depth);
+        srl_iterator_step_internal(iter, stack_ptr);
     }
 
     if (expect_false(n != 0)) {
@@ -500,85 +577,77 @@ srl_iterator_next(pTHX_ srl_iterator_t *iter, UV n)
 
 void
 srl_iterator_next_until_depth_and_idx(pTHX_ srl_iterator_t *iter, UV expected_depth, U32 expected_idx) {
-    U32 current_idx;
     srl_stack_t *stack = iter->stack;
+    srl_iterator_stack_ptr stack_ptr = stack->ptr;
     IV current_depth = SRL_STACK_DEPTH(stack);
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
-    SRL_ITER_TRACE("expected_depth=%"UVuf" expected_idx=%u",
-                   expected_depth, expected_idx);
-
     SRL_ITER_ASSERT_STACK(iter);
+
+    SRL_ITER_TRACE("expected_depth=%"UVuf" expected_idx=%u", expected_depth, expected_idx);
+    SRL_ITER_REPORT_STACK_STATE(iter);
+
+    if (expect_false((IV) expected_depth == current_depth && (IV) expected_idx == stack_ptr->idx))
+        return;
+
     if (expect_false((IV) expected_depth > current_depth)) {
         SRL_ITER_ERRORf2("srl_iterator_next_until_depth() can only go forward, "
                          "so expected_depth=%"UVuf" should not be greater then current_depth=%"IVdf,
                          expected_depth, current_depth);
     }
 
-    current_idx = stack->ptr->idx;
-    if (expect_false((IV) expected_depth == current_depth && expected_idx == current_idx))
-        return;
+    stack_ptr = stack->begin + expected_depth;
+    if (expect_false((IV) expected_idx > stack_ptr->idx)) {
+        SRL_ITER_ERRORf3("srl_iterator_next_until_depth() can only go forward, "
+                         "so expected_idx=%u should not be greater then current "
+                         "index (%u) at expected_depth=%"IVdf,
+                         expected_idx, stack_ptr->idx, expected_depth);
+    }
 
-    while (expect_true(!srl_stack_empty(stack))) {
-        srl_iterator_wrap_stack(aTHX_ iter, expected_depth);
+    stack_ptr = stack->ptr;
 
-        current_depth = SRL_STACK_DEPTH(stack);
-        if (expect_false(srl_stack_empty(stack)))
-            break;
+    while (1) {
+        srl_iterator_wrap_stack(iter, expected_depth, stack_ptr);
 
-        current_idx  = stack->ptr->idx;
-        if (current_depth == (IV) expected_depth && current_idx == expected_idx)
-            break;
-
-        if (expect_false(current_depth == (IV) expected_depth && expected_idx > current_idx)) {
-            SRL_ITER_ERRORf2("srl_iterator_next_until_depth() can only go forward, "
-                             "so expected_idx=%d should not be greater then current_idx=%d",
-                             expected_idx, current_idx);
+        if (SRL_STACK_DEPTH(stack) == (IV) expected_depth) {
+            if (stack_ptr->idx == (IV) expected_idx)  break;
+            assert(((IV) expected_idx > stack_ptr->idx) == 0);
         }
 
-        srl_iterator_step_internal(aTHX_ iter);
-    }
-
-    if (expect_false(current_depth != (IV) expected_depth)) {
-        SRL_ITER_ERRORf2("func led to wrong stack depth, expected=%"IVdf", actual=%"IVdf,
-                          expected_depth, current_depth);
-    }
-
-    if (expect_false(current_idx != expected_idx)) {
-        SRL_ITER_ERRORf2("func led to wrong stack index, expected=%u, actual=%u",
-                          expected_idx, current_idx);
+        srl_iterator_step_internal(iter, stack_ptr);
     }
 
+    assert(stack_ptr->idx == expected_idx);
+    assert(SRL_STACK_DEPTH(stack) == expected_depth);
     SRL_ITER_TRACE("Reached expected stack depth: %"UVuf " and idx: %u",
                    expected_depth, expected_idx);
+
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 }
 
 void
 srl_iterator_step_out(pTHX_ srl_iterator_t *iter, UV n)
 {
-    UV offset;
     srl_stack_t *stack = iter->stack;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
-    SRL_ITER_TRACE("n=%"UVuf, n);
-
     SRL_ITER_ASSERT_STACK(iter);
+
+    SRL_ITER_TRACE("n=%"UVuf, n);
+    SRL_ITER_REPORT_STACK_STATE(iter);
+
     // SRL_ITER_ASSERT_EOF(iter, "serialized object"); XXX need ability to go back on last element
     // if (expect_false(n == 0)) return; XXX keep it as a feature?
 
     while (n--) {
-        srl_stack_pop_nocheck(stack);
-        if (expect_false(srl_stack_empty(stack))) {
+        if (expect_false(SRL_ITER_STACK_ON_ROOT(stack))) {
             SRL_ITER_ERROR("It was last object on stack, no more parents");
         }
+
+        srl_stack_pop_nocheck(stack);
     }
 
-    offset = stack->ptr->offset; 
-    iter->buf.pos = iter->buf.body_pos + offset;
-    stack->ptr->idx = stack->ptr->count;
-
-    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
+    srl_iterator_restore_stack_position(aTHX_ iter);
 }
 
 IV
@@ -586,38 +655,40 @@ srl_iterator_array_goto(pTHX_ srl_iterator_t *iter, I32 idx)
 {
     I32 s_idx;
     srl_stack_t *stack = iter->stack;
+    srl_iterator_stack_ptr stack_ptr = stack->ptr;
 
     DEBUG_ASSERT_RDR_SANE(iter->pbuf);
-    SRL_ITER_TRACE("idx=%d", idx);
-
     SRL_ITER_ASSERT_EOF(iter, "array element");
     SRL_ITER_ASSERT_STACK(iter);
     SRL_ITER_ASSERT_ARRAY_ON_STACK(iter);
 
+    SRL_ITER_TRACE("idx=%d", idx);
+    SRL_ITER_REPORT_STACK_STATE(iter);
+
     if (idx >= 0) {
-        s_idx = stack->ptr->count - idx;
-        if (idx >= (I32) stack->ptr->count) {
-            SRL_ITER_TRACE("Index is out of range, idx=%d count=%u", idx, stack->ptr->count);
+        s_idx = stack_ptr->count - idx;
+        if (idx >= (I32) stack_ptr->count) {
+            SRL_ITER_TRACE("Index is out of range, idx=%d count=%u", idx, stack_ptr->count);
             return SRL_ITER_NOT_FOUND;
         }
     } else {
         s_idx = -idx;
-        if (s_idx > (I32) stack->ptr->count) {
-            SRL_ITER_TRACE("Index is out of range, idx=%d count=%u", idx, stack->ptr->count);
+        if (s_idx > (I32) stack_ptr->count) {
+            SRL_ITER_TRACE("Index is out of range, idx=%d count=%u", idx, stack_ptr->count);
             return SRL_ITER_NOT_FOUND;
         }
     }
 
-    if (s_idx == stack->ptr->idx) {
+    if (s_idx == stack_ptr->idx) {
         return SRL_RDR_BODY_POS_OFS(iter->pbuf); // already at expected position
-    } else if (s_idx > stack->ptr->idx) {
+    } else if (s_idx > stack_ptr->idx) {
         SRL_ITER_ERRORf2("Can't go backwards, idx=%d, count=%u",
-                         idx, stack->ptr->count);
+                         idx, stack_ptr->count);
     }
 
     // srl_iterator_next garantee that we remans on current stack
-    srl_iterator_next(aTHX_ iter, stack->ptr->idx - s_idx);
-    assert(stack->ptr->idx == s_idx);
+    srl_iterator_next(aTHX_ iter, stack_ptr->idx - s_idx);
+    assert(stack_ptr->idx == s_idx);
     return SRL_RDR_BODY_POS_OFS(iter->pbuf);
 }
 
@@ -630,12 +701,11 @@ srl_iterator_hash_key(pTHX_ srl_iterator_t *iter, STRLEN *len_out)
     srl_reader_char_ptr orig_pos = iter->buf.pos;
     *len_out = 0;
 
+    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_EOF(iter, "stringish");
     SRL_ITER_ASSERT_STACK(iter);
     SRL_ITER_ASSERT_HASH_ON_STACK(iter);
 
-    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
-
     tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;
     SRL_ITER_REPORT_TAG(iter, tag);
     iter->buf.pos++;
@@ -721,18 +791,20 @@ srl_iterator_hash_exists(pTHX_ srl_iterator_t *iter, const char *name, STRLEN na
 
     srl_stack_t *stack = iter->stack;
     IV stack_depth = SRL_STACK_DEPTH(stack);
+    srl_iterator_stack_ptr stack_ptr = stack->ptr;
 
+    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_ASSERT_EOF(iter, "stringish");
     SRL_ITER_ASSERT_STACK(iter);
     SRL_ITER_ASSERT_HASH_ON_STACK(iter);
 
-    DEBUG_ASSERT_RDR_SANE(iter->pbuf);
     SRL_ITER_TRACE("name=%.*s", (int) name_len, name);
+    SRL_ITER_REPORT_STACK_STATE(iter);
 
-    while (stack->ptr->idx) {
-        stack->ptr->idx--; // do not make it be part of while clause
+    while (stack_ptr->idx) {
+        stack_ptr->idx--; // do not make it be part of while clause
         SRL_ITER_ASSERT_STACK(iter);
-        assert(stack->ptr->idx % 2 == 1);
+        assert(stack_ptr->idx % 2 == 1);
         assert(SRL_STACK_DEPTH(stack) == stack_depth);
         DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 
@@ -807,6 +879,7 @@ srl_iterator_hash_exists(pTHX_ srl_iterator_t *iter, const char *name, STRLEN na
 
         // srl_iterator_next garantee that we remans on current stack
         srl_iterator_next(aTHX_ iter, 1);
+        stack_ptr = stack->ptr;
     }
 
     SRL_ITER_TRACE("didn't found key '%.*s'", (int) name_len, name);
@@ -930,6 +1003,10 @@ srl_iterator_stack_info(pTHX_ srl_iterator_t *iter, UV *length_ptr)
     if (length_ptr) *length_ptr = stack_ptr->count;
 
     switch (stack_ptr->tag) {
+        case SRL_ITER_STACK_ROOT_TAG:
+            type = SRL_ITERATOR_OBJ_IS_ROOT;
+            break;
+
         case SRL_HDR_HASH:
         CASE_SRL_HDR_HASHREF:
             type = SRL_ITERATOR_OBJ_IS_HASH;
@@ -963,4 +1040,3 @@ srl_iterator_decode(pTHX_ srl_iterator_t *iter)
     srl_decode_single_value(aTHX_ iter->dec, into, NULL);
     return into;
 }
-
